﻿/**
 * @license Copyright © 2013 onwards, Andrew Whewell
 * All rights reserved.
 *
 * Redistribution and use of this software in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 *    * Neither the name of the author nor the names of the program's contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OF THE SOFTWARE BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/**
 * @fileoverview A class that can auto-select aircraft based on a set of criteria.
 */

(function(VRS, $, undefined)
{
    //region Global settings
    VRS.globalOptions = VRS.globalOptions || {};
    VRS.globalOptions.aircraftAutoSelectEnabled = VRS.globalOptions.aircraftAutoSelectEnabled !== undefined ? VRS.globalOptions.aircraftAutoSelectEnabled : true;   // True if auto-select is enabled by default.
    VRS.globalOptions.aircraftAutoSelectClosest = VRS.globalOptions.aircraftAutoSelectClosest !== undefined ? VRS.globalOptions.aircraftAutoSelectClosest : true;   // True if the closest matching aircraft should be auto-selected, false if the furthest matching aircraft should be auto-selected.
    VRS.globalOptions.aircraftAutoSelectDeselectOffRadar = VRS.globalOptions.aircraftAutoSelectDeselectOffRadar !== undefined ? VRS.globalOptions.aircraftAutoSelectDeselectOffRadar : true;    // True if auto-select should de-select aircraft when they go off the radar.
    VRS.globalOptions.aircraftAutoSelectFilters = VRS.globalOptions.aircraftAutoSelectFilters || [];                                                                // The list of filters that auto-select starts off with.
    VRS.globalOptions.aircraftAutoSelectFiltersLimit = VRS.globalOptions.aircraftAutoSelectFiltersLimit !== undefined ? VRS.globalOptions.aircraftAutoSelectFiltersLimit : 2; // The maximum number of auto-select filters that we're going to allow.
    //endregion

    //region AircraftAutoSelect
    /**
     * A class that can auto-select aircraft on each refresh.
     * @param {VRS.AircraftList} aircraftList
     * @param {string=} name
     * @constructor
     */
    VRS.AircraftAutoSelect = function(aircraftList, name)
    {
        //region -- Fields
        var that = this;
        var _AircraftList = aircraftList;
        var _Dispatcher = new VRS.EventHandler({
            name: 'VRS.AircraftAutoSelect'
        });
        var _Events = {
            enabledChanged: 'enabledChanged'
        };
        //endregion

        //region -- Events subscribed, dispose
        var _AircraftListUpdatedHook = _AircraftList.hookUpdated(aircraftListUpdated, this);
        var _SelectedAircraftChangedHook = _AircraftList.hookSelectedAircraftChanged(selectedAircraftChanged, this);

        /**
         * Releases all of the resources held by the object.
         */
        this.dispose = function()
        {
            if(_AircraftListUpdatedHook) _AircraftList.unhook(_AircraftListUpdatedHook);
            if(_SelectedAircraftChangedHook) _AircraftList.unhook(_SelectedAircraftChangedHook);
            _AircraftListUpdatedHook = _SelectedAircraftChangedHook = null;
        };
        //endregion

        //region -- Properties
        var _Name = name || 'default';
        /**
         * Gets the name of the object. This is used when storing and loading settings.
         * @returns {string}
         */
        this.getName = function() { return _Name; };

        /**
         * Determines whether the object is actively setting selected aircraft.
         * @type {bool}
         * @private
         */
        var _Enabled = VRS.globalOptions.aircraftAutoSelectEnabled;
        /**
         * Gets a value indicating whether the object is actively setting the selected aircraft.
         * @returns {bool}
         */
        this.getEnabled = function() { return _Enabled; };
        /**
         * Sets a value indicating whether the object is to auto-select aircraft or not.
         * @param {bool} value
         */
        this.setEnabled = function(value) {
            if(_Enabled !== value) {
                _Enabled = value;
                _Dispatcher.raise(_Events.enabledChanged);

                if(_Enabled && _AircraftList) {
                    var closestAircraft = that.closestAircraft(_AircraftList);
                    if(closestAircraft && closestAircraft !== _AircraftList.getSelectedAircraft()) {
                        _AircraftList.setSelectedAircraft(closestAircraft, false);
                    }
                }
            }
        };

        /**
         * True if the closest aircraft is to be selected, false otherwise.
         * @type {bool}
         * @private
         */
        var _SelectClosest = VRS.globalOptions.aircraftAutoSelectClosest;
        /**
         * Gets a value indicating that the closest aircraft is to be selected.
         * @returns {boolean}
         */
        this.getSelectClosest = function() { return _SelectClosest; };
        /**
         * Sets a value indicating that the closest aircraft is to be selected.
         * @param value
         */
        this.setSelectClosest = function(value) { _SelectClosest = value; };

        /**
         * True if selected aircraft are to be deselected when they are no longer in range.
         * @type {bool}
         * @private
         */
        var _DeselectWhenOffRadar = VRS.globalOptions.aircraftAutoSelectDeselectOffRadar;
        /**
         * Gets a value indicating that the selected aircraft is to be deselected when it goes out of range.
         * @returns {boolean}
         */
        this.getDeselectWhenOffRadar = function() { return _DeselectWhenOffRadar; };
        /**
         * Sets a value indicating that the selected aircraft is to be deselected when it goes out of range.
         * @param value
         */
        this.setDeselectWhenOffRadar = function(value) { _DeselectWhenOffRadar = value; };

        /**
         * An array of the different aircraft filters to use when determining which aircraft to auto-select.
         * @type {VRS.AircraftFilter[]}
         * @private
         */
        var _Filters = VRS.globalOptions.aircraftAutoSelectFilters;
        //noinspection JSUnusedGlobalSymbols
        /**
         * Returns a clone of the filter at the specified index.
         * @param {number} index
         * @returns {VRS.AircraftFilter}
         */
        this.getFilter = function(index) { return /** @type {VRS.AircraftFilter} */ _Filters[index].clone(); };
        //noinspection JSUnusedGlobalSymbols
        /**
         * Sets the property filter at the specified index.
         * @param {number}              index
         * @param {VRS.AircraftFilter}  aircraftFilter
         */
        this.setFilter = function(index, aircraftFilter) {
            if(_Filters.length < VRS.globalOptions.aircraftListFiltersLimit) {
                var existing = index < _Filters.length ? _Filters[index] : null;
                if(existing && !existing.equals(aircraftFilter)) _Filters[index] = aircraftFilter;
            }
        };
        //endregion

        //region -- Events exposed
        /**
         * Raised after the Enabled property value has changed.
         * @param {function} callback
         * @param {object} forceThis
         * @returns {object}
         */
        this.hookEnabledChanged = function(callback, forceThis) { return _Dispatcher.hook(_Events.enabledChanged, callback, forceThis); };

        /**
         * Unhooks an event hooked on this object.
         * @param {object} hookResult
         * @returns {*}
         */
        this.unhook = function(hookResult) { return _Dispatcher.unhook(hookResult); };
        //endregion

        //region -- State persistence - saveState, loadState, applyState, loadAndApplyState
        /**
         * Saves the current state of the object.
         */
        this.saveState = function()
        {
            VRS.configStorage.save(persistenceKey(), createSettings());
        };

        /**
         * Returns the saved state of the object.
         * @returns {VRS_STATE_AIRCRAFTAUTOSELECT}
         */
        this.loadState = function()
        {
            var savedSettings = VRS.configStorage.load(persistenceKey(), {});
            var result = $.extend(createSettings(), savedSettings);
            result.filters = VRS.aircraftFilterHelper.buildValidFiltersList(result.filters, VRS.globalOptions.aircraftAutoSelectFiltersLimit);

            return result;
        };

        /**
         * Applies a saved state to the object.
         * @param {VRS_STATE_AIRCRAFTAUTOSELECT} settings
         */
        this.applyState = function(settings)
        {
            that.setEnabled(settings.enabled);
            that.setSelectClosest(settings.selectClosest);
            that.setDeselectWhenOffRadar(settings.deselectOffRadar);

            _Filters = [];
            $.each(settings.filters, function(idx, serialisedFilter) {
                var aircraftFilter = VRS.aircraftFilterHelper.createFilter(serialisedFilter.property);
                aircraftFilter.applySerialisedObject(serialisedFilter);
                _Filters.push(aircraftFilter);
            });
        };

        /**
         * Loads and applies the saved state to the object.
         */
        this.loadAndApplyState = function()
        {
            that.applyState(that.loadState());
        };

        /**
         * Returns the key under which the object's settings are saved.
         * @returns {string}
         */
        function persistenceKey()
        {
            return 'vrsAircraftAutoSelect-' + _Name;
        }

        /**
         * Creates the state object.
         * @returns {VRS_STATE_AIRCRAFTAUTOSELECT}
         */
        function createSettings()
        {
            var filters = VRS.aircraftFilterHelper.serialiseFiltersList(_Filters);
            return {
                enabled: _Enabled,
                selectClosest: _SelectClosest,
                deselectOffRadar: _DeselectWhenOffRadar,
                filters: filters
            };
        }
        //endregion

        //region -- createOptionPane
        /**
         * Creates the option pane that allows configuration of the object.
         * @param {number} displayOrder
         * @returns {VRS.OptionPane}
         */
        this.createOptionPane = function(displayOrder)
        {
            var pane = new VRS.OptionPane({
                name:           'vrsAircraftAutoSelectPane',
                titleKey:       'PaneAutoSelect',
                displayOrder:   displayOrder
            });
            if(VRS.aircraftListFiltersLimit !== 0) {
                pane.addField(new VRS.OptionField({
                    name:           'deselect',
                    controlType:    VRS.optionControlTypes.checkBox,
                    labelKey:       'DeselectWhenOffRadar',
                    getValue:       that.getDeselectWhenOffRadar,
                    setValue:       that.setDeselectWhenOffRadar,
                    saveState:      that.saveState
                }));

                pane.addField(new VRS.OptionField({
                    name:           'enable',
                    controlType:    VRS.optionControlTypes.checkBox,
                    labelKey:       'AutoSelectAircraft',
                    getValue:       that.getEnabled,
                    setValue:       that.setEnabled,
                    saveState:      that.saveState,
                    hookChanged:    that.hookEnabledChanged,
                    unhook:         that.unhook
                }));

                pane.addField(new VRS.OptionField({
                    name:           'closest',
                    controlType:    VRS.optionControlTypes.radioButton,
                    labelKey:       'Select',
                    getValue:       that.getSelectClosest,
                    setValue:       that.setSelectClosest,
                    saveState:      that.saveState,
                    typeSettings:   new VRS.OptionRadioButtonTypeSettings({ values: [
                        new VRS.ValueText({ value: true, textKey: 'ClosestToCurrentLocation' }),
                        new VRS.ValueText({ value: false, textKey: 'FurthestFromCurrentLocation' })
                    ]})
                }));

                var panesTypeSettings = VRS.aircraftFilterHelper.addConfigureFiltersListToPane({
                    pane:       pane,
                    filters:    _Filters,
                    saveState:  function() { that.saveState(); },
                    maxFilters: VRS.globalOptions.aircraftAutoSelectFiltersLimit,
                    allowableProperties: [
                        VRS.AircraftFilterProperty.Altitude,
                        VRS.AircraftFilterProperty.Distance
                    ],
                    addFilter: function(newComparer, paneSettings) {
                        var aircraftFilter = that.addFilter(newComparer);
                        paneSettings.addPane(aircraftFilter.createOptionPane(that.saveState));
                        that.saveState();
                    },
                    addFilterButtonLabel: 'AddCondition'
                });

                panesTypeSettings.hookPaneRemoved(filterPaneRemoved, this);
            }

            return pane;
        };
        //endregion

        //region --addFilter, comparerPaneRemoved, removeFilterAt
        /**
         * Adds a new filter to the auto-select settings.
         * @param {VRS.AircraftFilter|VRS.AircraftFilterProperty} filterOrFilterProperty
         * @returns {VRS.AircraftFilter}
         */
        this.addFilter = function(filterOrFilterProperty)
        {
            var filter = filterOrFilterProperty;
            if(!(filter instanceof VRS.AircraftFilter)) {
                filter = VRS.aircraftFilterHelper.createFilter(filterOrFilterProperty);
            }
            _Filters.push(filter);

            return filter;
        };

        //noinspection JSUnusedLocalSymbols
        /**
         * Removes an existing filter from the auto-select settings.
         * @param {VRS.OptionPane} [pane] Unused.
         * @param {number} index
         */
        function filterPaneRemoved(pane, index)
        {
            that.removeFilterAt(index);
            that.saveState();
        }

        /**
         * Removes an existing filter from the auto-select settings.
         * @param {number} index
         */
        this.removeFilterAt = function(index)
        {
            _Filters.splice(index, 1);
        };
        //endregion

        //region -- closestAircraft, aircraftListUpdated, selectedAircraftChanged
        /**
         * Returns the closest aircraft from the aircraft list passed across.
         * @param {VRS.AircraftList} aircraftList
         * @returns {VRS.Aircraft}
         */
        this.closestAircraft = function(aircraftList)
        {
            var result = null;

            if(this.getEnabled()) {
                var useClosest = this.getSelectClosest();
                var useFilters = _Filters.length > 0;
                var selectedDistance = useClosest ? Number.MAX_VALUE : Number.MIN_VALUE;
                aircraftList.foreachAircraft(function(aircraft) {
                    if(aircraft.distanceFromHereKm.val !== undefined) {
                        var isCandidate = useClosest ? aircraft.distanceFromHereKm.val < selectedDistance : aircraft.distanceFromHereKm.val > selectedDistance;
                        if(isCandidate) {
                            if(useFilters && !VRS.aircraftFilterHelper.aircraftPasses(aircraft, _Filters, {})) isCandidate = false;
                            if(isCandidate) {
                                result = aircraft;
                                selectedDistance = aircraft.distanceFromHereKm.val;
                            }
                        }
                    }
                });
            }

            return result;
        };

        //noinspection JSUnusedLocalSymbols
        /**
         * Called when the aircraft list is updated.
         * @param {VRS.AircraftCollection} newAircraft
         * @param {VRS.AircraftCollection} offRadar
         */
        function aircraftListUpdated(newAircraft, offRadar)
        {
            var selectedAircraft = _AircraftList.getSelectedAircraft();
            var autoSelectAircraft = that.closestAircraft(_AircraftList);

            if(autoSelectAircraft) {
                if(autoSelectAircraft !== selectedAircraft) _AircraftList.setSelectedAircraft(autoSelectAircraft, false);
            } else if(selectedAircraft && that.getDeselectWhenOffRadar()) {
                if(offRadar.findAircraftById(selectedAircraft.id)) _AircraftList.setSelectedAircraft(undefined, false);
            }
        }

        /**
         * Called when the aircraft list changes the selected aircraft.
         //* @param {VRS.Aircraft} oldSelectedAircraft
         */
        function selectedAircraftChanged()
        {
            if(_AircraftList.getWasAircraftSelectedByUser()) that.setEnabled(false);
        }
        //endregion
    };
    //endregion
}(window.VRS = window.VRS || {}, jQuery));